/*
 * ConvRBM.cuh
 *
 *  Created on: 27-Apr-2009
 *      Author: spoon
 */

#ifndef CONVRBM_H_
#define CONVRBM_H_

#include <matrix.h>
#include <nvmatrix.cuh>
#include <thread.h>
#include <queue.h>
#include <WorkRequest.h>
#include <WorkResult.h>

#define SPARSE_DAMPING 0.6
#define L1_WC false // use L1 weight decay? otherwise L2

#define WEIGHT_NORMALIZATION false

class ConvRBM : public Thread {
private:
    // ===========================
    // Host matrices
    // ===========================
    Matrix* hHidVis, *hHidVisInc;
    Matrix* hVisBiases, *hVisBiasInc;
    Matrix* hHidBiases, *hHidBiasInc;
    Matrix* hHidMeans;

    Matrix* hVisHidGlob, *hVisHidIncGlob;
    Matrix* hHidBiasesGlob, *hHidBiasIncGlob;
    Matrix* hHidMeansGlob;

    Matrix* hNegDataPCD;

    // ===========================
    // Device matrices
    // ===========================
    NVMatrix* data;
    NVMatrix* hidVis, *hidVisInc;
    NVMatrix* visBiases, *visBiasInc;
    NVMatrix* hidBiases, *hidBiasInc;
    NVMatrix* hidMeans;

    NVMatrix* visHidGlob, *visHidIncGlob, *visHidGlob_T;
    NVMatrix* hidBiasesGlob, *hidBiasIncGlob;
    NVMatrix* hidMeansGlob;

    // ---------------------------
    // Temporary storage
    // ---------------------------
    NVMatrix* hidProbs, *hidProbsPadded, *hidProbsSampled;
    NVMatrix* posUnitProds, *negUnitProds, *unitProds_tmp;
    NVMatrix* posVisAct, *negVisAct;
    NVMatrix* posHidAct, *negHidAct;
    NVMatrix* negDataFull, *negDataCenter, *reconData;
    NVMatrix* sparseGrads_tmp;
    NVMatrix* sparseBiasGrads, *sparseWeightGrads;
    NVMatrix* localRegions_tmp, *localMax_tmp, *localSum_tmp, *sampleRandoms;
    NVMatrix* hidProbs_tmp, *hidBiases_tiled;

    NVMatrix* miniDataZeroPadded;

    NVMatrix* hidProbsGlob;
    NVMatrix* posHidActGlob, *negHidActGlob;
    NVMatrix* negDataGlob;
    NVMatrix* sparseGradsGlob, *sparseWeightGradsGlob, *sparseBiasGradsGlob;

    // ===========================
    // Various learning parameters
    // ===========================
    float epsHidVis, epsVisBias, epsHidBias;
    float epsHidVisGlob, epsHidBiasGlob;
    float mom, weightCost;
    float sparseCost, sparseTarget;
    float sparseCostGlob, sparseTargetGlob;
    int minibatchSize, cd;

    // ===========================
    // RBM dimensions
    // ===========================
    int numVis, numHid, numConvOutputs, numConvOutputsX;
    int imgSize, imgCenterSize, filterSize, fixedBorderSize;
    int hidProbsPaddingSize, paddedHidProbsSize;
    int localMaxSize, localMaxPixels, zeroPadding, paddedImgSize;
    bool tiedBiases;
    int numHidGlob;

    // ================================
    // Communication with python thread
    // ================================
    Queue<WorkRequest*>* requestQueue;
    Queue<WorkResult*>* resultQueue;

    // Cuda device ID to run on
    int deviceID;

    void initCuda();
    void train();
    void test();
    void syncWithHost();
    void computeHidProbs(NVMatrix* miniData);
    void computeCDGrads(NVMatrix* miniData, bool posPhase);
    void computeBiasGrads(NVMatrix* hidAct, NVMatrix* target);
    void computeNegData(NVMatrix* posMiniData);
    double computeError(NVMatrix* pos);
    void updateWeights();
protected:
    void* run();
public:
    ConvRBM(Matrix* hHidVis, Matrix* hVisBiases, Matrix* hHidBiases,
            Matrix* hHidVisInc, Matrix* hVisBiasInc, Matrix* hHidBiasInc,
            Matrix* hHidMeans,
            Matrix* hVisHidGlob, Matrix *hHidBiasesGlob,
            Matrix* hVisHidIncGlob, Matrix* hHidBiasIncGlob,
            Matrix* hHidMeansGlob,
            Matrix* hNegDataPCD,
            int fixedBorderSize, int localMaxSize, int zeroPadding,
            float epsHidVis, float epsVisBias, float epsHidBias,
            float epsHidVisGlob, float epsHidBiasGlob,
            float mom, float weightCost,
            float sparseCost, float sparseTarget,
            float sparseCostGlob, float sparseTargetGlob,
            int minibatchSize, int cd,
            int deviceID);

    virtual ~ConvRBM();

    inline Queue<WorkRequest*>* getWorkRequestQueue() {
        return requestQueue;
    }

    inline Queue<WorkResult*>* getWorkResultQueue() {
        return resultQueue;
    }
};

#endif /* CONVRBM_H_ */
