import numpy as n
import os
from util import *

BATCH_META_FILE = "batches.meta"

class DataProvider:
    dp_types = {"default": "The default data provider, loads one batch into memory at a time",
                "cropped-image-28": "Same as default, but crops 32x32 iamges to 28x28",
                "memory": "Loads the entire dataset into memory",
                "labeled": "Returns data and labels (used by classifiers)",
                "noisy": "Data + noise provider (used by autoencoder)"}
    def __init__(self, data_dir, batch_range, init_epoch=None, init_batchnum=None):
        if init_epoch is None:
            init_epoch = 1
        if init_batchnum is None:
            init_batchnum = batch_range[0]
        self.data_dir = data_dir
        self.batch_range = batch_range
        self.curr_epoch = init_epoch
        self.curr_batchnum = init_batchnum
        self.batch_meta = self.get_batch_meta(data_dir)
        self.data_dic = None
        self.data_mean = None
        if "data_mean" in self.batch_meta:
            self.data_mean = self.batch_meta["data_mean"]

    def get_next_batch(self, dic=False):
        if self.data_dic is None or len(self.batch_range) > 1:
            self.data_dic = self.get_batch(self.curr_batchnum)
        epoch, batchnum = self.curr_epoch, self.curr_batchnum
        self.advance_batch()
        if dic:
            return epoch, batchnum, (self.data_dic,)
        return epoch, batchnum, (self.data_dic["data"],)
    
    def get_batch(self, batch_num):
        dic = unpickle(os.path.join(self.data_dir, self.__get_data_file_name(batch_num)))
        if self.data_mean is not None:
            dic['data'] -= self.data_mean
        return dic
    
    def get_data_dims(self):
        return self.batch_meta['num_vis']
    
    def advance_batch(self):
        if self.curr_batchnum == self.batch_range[-1]:
            self.curr_epoch += 1
            self.curr_batchnum = self.batch_range[0]
        else:
            self.curr_batchnum += 1
            
    # private
    def __get_data_file_name(self, batchnum=None):
        if batchnum is None:
            batchnum = self.curr_batchnum
        return 'data_batch_%d' % batchnum
    
    @staticmethod
    def get_instance(data_dir, batch_range, init_epoch=None, init_batchnum=None, type="default"):
        if type == "default":
            return DataProvider(data_dir, batch_range, init_epoch, init_batchnum)
        elif type == "cropped-image-28":
            return CroppedImageDataProvider(data_dir, batch_range, init_epoch, init_batchnum)
        elif type == "memory":
            return MemoryDataProvider(data_dir, batch_range, init_epoch, init_batchnum)
        elif type == "labeled":
            return LabeledDataProvider(data_dir, batch_range, init_epoch, init_batchnum)
        elif type == "noisy":
            return NoisyDataProvider(data_dir, batch_range, init_epoch, init_batchnum)
        raise DataProviderException("No such data provider: %s" % type)
    
    @staticmethod
    def get_batch_meta(data_dir):
        return unpickle(os.path.join(data_dir, BATCH_META_FILE))
    
    @staticmethod
    def get_num_batches(srcdir):
        m = re.compile('^data_batch_\d+$')
        return len([f for f in os.listdir(srcdir) if m.match(f)])

class CroppedImageDataProvider(DataProvider):
    def get_next_batch(self, dic=False):
        epoch, batchnum, _ = DataProvider.get_next_batch(self)
        
        if dic:
            self.data_dic['data'] = CroppedImageDataProvider.__trim_borders(self.data_dic['data'])
            return epoch, batchnum, (self.data_dic,)

        return epoch, batchnum, (CroppedImageDataProvider.__trim_borders(self.data_dic['data']),)
    
    def get_data_dim(self):
        return 28*28*3
        
    # private
    @staticmethod
    def __trim_borders(x):
        y = x.reshape(x.shape[0], 3,32,32)
        y = y[:,:,2:30,2:30]
        return n.array(y.reshape(x.shape[0], 28*28*3), dtype=n.single)

class MemoryDataProvider(DataProvider):
    def __init__(self, data_dir, batch_range, init_epoch=None, init_batchnum=None):
        DataProvider.__init__(self, data_dir, batch_range, init_epoch, init_batchnum)
        self.data_dic = []
        for i in batch_range:
            self.data_dic += [unpickle(os.path.join(self.data_dir, self.__get_data_file_name(i)))]
    
    def get_next_batch(self, dic=False):
        epoch, batchnum = self.curr_epoch, self.curr_batchnum
        self.advance_batch()
        if dic:
            return epoch, batchnum, (self.data_dic[batchnum - self.batch_range[0]],)
        return epoch, batchnum, (self.data_dic[batchnum - self.batch_range[0]]["data"],)

class LabeledDataProvider(DataProvider):
    def get_next_batch(self, dic=False):
        epoch, batchnum, _ = DataProvider.get_next_batch(self)
        
        if dic:
            return epoch, batchnum, (self.data_dic,)
        return epoch, batchnum, (self.data_dic['data'], n.c_[n.array(self.data_dic['labels'], dtype=n.single)])
    
    def get_num_categories(self):
        return len(self.batch_meta['label_names'])

class NoisyDataProvider(DataProvider):
    def __init__(self, data_dir, batch_range, init_epoch=None, init_batchnum=None):
        DataProvider.__init__(self, data_dir[0], batch_range, init_epoch, init_batchnum)
        self.noise_dir = data_dir[1]
        self.noise_dic = None
        
    def get_next_batch(self, dic=False):
        epoch, batchnum, _ = DataProvider.get_next_batch(self)
        if self.noise_dir:
            self.noise_dic = unpickle(os.path.join(self.noise_dir, self.__get_data_file_name()))
        if dic:
            return epoch, batchnum, (self.data_dic, self.noise_dic)
        noise = self.noise_dic and self.noise_dic["noise"] or None
        return epoch, batchnum, (self.data_dic["data"], noise)
    
class DataProviderException(Exception):
    pass
